// Navigator.cc
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stream.h>
#include <sys/time.h>

#include <Performer/pf/pfDCS.h>
#include <Performer/pfdu.h>
#include <pfcave.h>

#include "NavBase.hh"
#include "NavPath.hh"
#include "Navigator.hh"
#include "CavePointer.hh"
//#define DEFAULT_TSPEED 2.0f/1.0e6
//#define DEFAULT_RSPEED 1.0f/1.0e6
#define DEFAULT_TSPEED 30.0f
#define DEFAULT_RSPEED 30.0f
#ifndef PI
#define PI 3.1415
#endif

#define THRESH(val,thresh) (val>thresh || val<-thresh)

//==========>Navigator<===============================
Navigator::Navigator(pfDCS *nav,pfNode *object):
  navdcs(nav),modeldcs(object),collision_detection(0),
  modelnav(0),navstate(0),navhalt(0){
    //NavBase initnav(navdcs);
    //initnav.loadIdentity(); // initialize
    setMode(Off);
    rSpeed=DEFAULT_RSPEED;
    tSpeed=DEFAULT_TSPEED;
    navstate->loadIdentity(); // initialize nav
    navstate->updateDCS();
}

Navigator::Navigator(pfDCS *nav):
  navdcs(nav),modeldcs(0),collision_detection(0),
  modelnav(0),navstate(0),navhalt(0){
    //NavBase initnav(navdcs);
    //initnav.loadIdentity(); // initialize
  setMode(Off); // equiv to NavNull
  rSpeed=DEFAULT_RSPEED;
  tSpeed=DEFAULT_TSPEED;
  navstate->loadIdentity(); // initialize nav
  navstate->updateDCS();
}

void Navigator::update(){
  //cerr << "Navigator::update()" << endl;
  if(navhalt>0) return;
  navstate->update();
  if(!modelnav) // KLUDGE!!
    navstate->updateDCS();
  // should we commit the navdcs here?
  // or do it in the subclass?
}

int Navigator::update(NavMode nav){
  int retval=setMode(nav);
  update();
  return retval;
}

int Navigator::setMode(NavMode nav){
  if(nav==Halt){   
    navhalt++;
    cerr << "setmode HALT!, navhalt=" << navhalt << endl;
    return 0;
  }
  else if(nav==Continue){
    navhalt--;
    cerr << "setmode CONTINUE!, navhalt=" << navhalt << endl;
    return 0;
  }
  if(nav==Reset){
    cerr << "Navigation RESET!!!" << endl;
    navstate->reset(); // should load identity
    return 0; // didn't change mode
  }
  else if(nav!=mode){
    if(navstate) delete navstate;
    modelnav=0;
    mode=nav;
    //cerr << "ChangeMode==================" << endl;
    switch(nav){
    case GrabModelAbsolute:
      /*
	if(modeldcs){
	modelnav=1;
	navstate=new NavGrabAbsolute(modeldcs);
	break;
	}// else fallthrough
	*/
    case GrabAbsolute:
      navstate = new NavGrabAbsolute(navdcs);
      break;
    case GrabModelRelative:
      /*
	if(modeldcs){
	modelnav=1;
	navstate=new NavGrabRelative(modeldcs);
	break;
	}// otherwise, fallthrough
	*/
    case GrabRelative:
      navstate = new NavGrabRelative(navdcs);
      break;     
    case Walk:
      if(modeldcs){
	navstate = new NavWalk(navdcs,modeldcs);
	((NavWalk*)navstate)->setRotationSpeed(rSpeed);
	((NavWalk*)navstate)->setTranslationSpeed(tSpeed);
	break;
      }
      // else fall-thru
    case Fly:
      navstate = new NavFly(navdcs);
      ((NavFly*)navstate)->setRotationSpeed(rSpeed);
      ((NavFly*)navstate)->setTranslationSpeed(tSpeed);
      break;
    case Path:
      cerr << "Switching to NavPath" << endl;
      navstate = new NavPath(navdcs);
      // note on instantiation, position will snap to track...
      break;
    case Off: 
    default:
      navstate=new NavBase(navdcs);
      break;
    }
    return 1; // mode changed
  }
  else 
    return 0; // no mode change
}

void Navigator::collisionDetection(int val){
  collision_detection=val;
  if(!modeldcs) collision_detection=0;
}

void Navigator::setRotationSpeed(float speed){
  // where speed is in rad/sec
  rSpeed=speed;
  if(mode==Walk || mode==Fly)
    ((NavFly*)navstate)->setRotationSpeed(rSpeed);
}

void Navigator::adjustRotationSpeed(float factor){
  rSpeed*=factor;
  if(mode==Walk || mode==Fly)
    ((NavFly*)navstate)->setRotationSpeed(rSpeed);
}

void Navigator::setTranslationSpeed(float speed){
  // where speed is the trans speed in clicks/sec
  tSpeed=speed;
  if(mode==Walk || mode==Fly)
    ((NavFly*)navstate)->setTranslationSpeed(tSpeed);
}

void Navigator::adjustTranslationSpeed(float factor){
  tSpeed*=factor;
  if(mode==Walk || mode==Fly)
    ((NavFly*)navstate)->setTranslationSpeed(tSpeed);
}

int Navigator::callback(ParamSet *caller,ParamSet *data){
  Param *p;
  if(!data)
    return 0;
  if((p=data->get("setMode"))){
    char *modename;
    if(p->isOfType(String) && p->data && (modename=(char *)(p->data))){ 
      // lets just use strings for messages for now
      if(!strcmp(modename,"Walk"))
	setMode(Walk);
      else if(!strcmp(modename,"Fly"))
	setMode(Fly);
      else if(!strcmp(modename,"Reset"))
	setMode(Reset);
      else if(!strcmp(modename,"Halt"))
	setMode(Halt);
      else if(!strcmp(modename,"Path"))
	setMode(Path);
      else if(!strcmp(modename,"Continue"))
	setMode(Continue);
    }
  }
  caller=0;
  return 0;
}

